/*
   Bone Animation 2 Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to perform bone animation against a set of 2D quads.
   We will build off of the first bone animation demo and this time allow the bones
   to manipulate a set of geometry.
*/


// Demo: Bone Animation 2.


#define WIN32_LEAN_AND_MEAN
#define VC_LEANMEAN

#include<windows.h>
#include<d3dx9.h>
#include"CVector.h"
#include"CBone.h"
#include"CMatrix.h"

#define D3DFVF_D3DVertex (D3DFVF_XYZ | D3DFVF_DIFFUSE)

// Function Prototypes...
bool InitializeDirect3D(HWND hwnd, bool fullscreen);
bool InitializeObject();
void RenderScene();
void ShutdownDirect3D();

// Here are our Direct3D objects.
LPDIRECT3D9 Direct3D_Object = NULL;
LPDIRECT3DDEVICE9 D3D_Device = NULL;

// Projection matrix.
D3DXMATRIXA16 Projection;

// Vertex stucture.
struct Vertex
{
   FLOAT x, y, z;
   DWORD color;
};

#define MAX_BONES                2
#define MAX_MESHES               3
#define MAX_MESH_POINTS          4

// xRotation and yRotation is used to rotate bones.
float xRotation = 0.0f;
float yRotation = 0.0f;

// Used to move our model with the arrow keys.
CVector4 trans;

// Hierarchy of bones kept in a simple array instead of a hierarchy data structure.
CBone bones[MAX_BONES];

// We will use this set of matrices to help us draw out the bones.
CMatrix4x4 displayBones[MAX_BONES];

// List of vertex points of the model.  The model is made up of 3 quads with 4 points each.
CBoneVertex modelPoints[MAX_MESHES * MAX_MESH_POINTS];


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   static POINT oldMousePos;              // Last mouse position.
   static POINT currentMousePos;          // Current mouse position.
   static bool isMouseActive;             // Is the left mouse button down.

   switch(message)
      {
	   case WM_KEYUP:
         // If the user presses the escape key then exit the application.
         if(wParam == VK_ESCAPE)
            PostQuitMessage(0);
         break;

      case WM_KEYDOWN:
		   switch(wParam) 
		      {
               case VK_UP:
                  trans.z -= 0.5f;
                  break;

               case VK_DOWN:
                  trans.z += 0.5f;
                  break;

               case VK_LEFT:
                  trans.x += 0.5f;
                  break;

               case VK_RIGHT:
                  trans.x -= 0.5f;
                  break;
		      }
         break;

      case WM_LBUTTONDOWN:
				oldMousePos.x = currentMousePos.x = LOWORD (lParam);
				oldMousePos.y = currentMousePos.y = HIWORD (lParam);
            isMouseActive = true;

			   break;


			case WM_LBUTTONUP:
				isMouseActive = false;

				break;


			case WM_MOUSEMOVE:
				currentMousePos.x = LOWORD (lParam);
				currentMousePos.y = HIWORD (lParam);

				if(isMouseActive)
					{
                  xRotation -= (currentMousePos.x - oldMousePos.x);
						yRotation -= (currentMousePos.y - oldMousePos.y);
					}

				oldMousePos.x = currentMousePos.x;
				oldMousePos.y = currentMousePos.y;

			   break;

      case WM_DESTROY:
      case WM_CLOSE:
         PostQuitMessage(0);
         break;

         default:
            break;
      }

   return DefWindowProc(hwnd, message, wParam, lParam);
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                                           // Message object.
   HWND hwnd;                                         // Handle to the window.
   WNDCLASSEX windowClass;                            // Window class object.
   bool done = false;                                 // true if we should exit.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);           // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;       // style of the window.
   windowClass.lpfnWndProc = WndProc;                 // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                        // Extra class information.
   windowClass.cbWndExtra = 0;                        // Extra window information.
   windowClass.hInstance = hInstance;                 // Handle of application Instance.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);// Window Icon.
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW); // Load mouse cursor.
   windowClass.hbrBackground = NULL;                  // Background color.
   windowClass.lpszMenuName = NULL;                   // Menu.
   windowClass.lpszClassName = "UGPClass";            // Name of the window class.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);// Minimized window icon.

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // After your class has been registered then you are ready to create your window.
   hwnd = CreateWindowEx(NULL,                                       // The extended style.
                         "UGPClass",                                 // Window class.
                         "Simple Bone Animation 2 - by The Programming Ace",     // Window name.
                          WS_OVERLAPPEDWINDOW | WS_VISIBLE |         // Window style.
                          WS_SYSMENU |WS_CLIPCHILDREN |              // Window style.
                          WS_CLIPSIBLINGS,                           // Window style.
                          100, 100,                                  // X, Y coords.
                          640, 480,                                  // Window size.
                          NULL,                                      // Handle to parent window.
                          NULL,                                      // Menu.
                          hInstance,                                 // Handle to app instance.
                          NULL);                                     // Pointer to window.

   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;

   ShowWindow(hwnd, SW_SHOW);    // Show the window.
   UpdateWindow(hwnd);           // Update its display.

   done = false;                 // false = run program, true means stop.

   // Initialize Direct3D.  If fails then we don't want to run the application.
   if(!InitializeDirect3D(hwnd, false)) done = true;

   // Application loop.
   while(!done)
      {
         if(PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     done = true;
                  }

               TranslateMessage(&msg);
               DispatchMessage(&msg);
            }
         else
            {
               RenderScene();
            }
      }

    // Here we shutdown Direct 3D.
	ShutdownDirect3D();

   // Here we unregister the window class with the OS.
   UnregisterClass("UGPClass", windowClass.hInstance);

   return (int)msg.wParam;
}


bool InitializeDirect3D(HWND hwnd, bool fullscreen)
{
   // This object will allow us to set the display mode of the screen.
   D3DDISPLAYMODE DisplayMode;
   
   // This is used to get the capabilities of the hardware.
   D3DCAPS9 D3DCaps;

   // This will allow us to set the parameters of the screen.
   D3DPRESENT_PARAMETERS Present_Parameters;
   ZeroMemory(&Present_Parameters, sizeof(Present_Parameters));

   // Create the Direct3D object to get everything started.
   Direct3D_Object = Direct3DCreate9(D3D_SDK_VERSION);
   if(Direct3D_Object == NULL) return false;

   // This function will get the display mode of the device and place it in DisplayMode.
   if(FAILED(Direct3D_Object->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &DisplayMode)))
      return false;

   // Get the capabilities of the hardware.
	if(FAILED(Direct3D_Object->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &D3DCaps)))
      return false;

   // Test which is supported, hardware or software vertex processing.
	DWORD VProcess = 0;

	if(D3DCaps.VertexProcessingCaps != 0) VProcess |= D3DCREATE_HARDWARE_VERTEXPROCESSING;
	else VProcess |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;

   // Here we are setting the applications parameters...
   if(fullscreen)
      {
         Present_Parameters.Windowed = FALSE;               // Window mode (fullscreen).
         Present_Parameters.BackBufferWidth = DisplayMode.Width;
         Present_Parameters.BackBufferHeight = DisplayMode.Height;
      }
   else
      Present_Parameters.Windowed = TRUE;                   // Window mode (not fullscreen).
   Present_Parameters.SwapEffect = D3DSWAPEFFECT_DISCARD;   // Dealing with animation (see doc).
   Present_Parameters.BackBufferFormat = DisplayMode.Format;// Render to the area of the screen.
   Present_Parameters.BackBufferCount = 1;                  // Number of back buffers.
   Present_Parameters.EnableAutoDepthStencil = TRUE;        // Check documentation.
   Present_Parameters.AutoDepthStencilFormat = D3DFMT_D16;  // Check documentation.

   // Now we must create the rendering device.
   if(FAILED(Direct3D_Object->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hwnd,
                     VProcess, &Present_Parameters, &D3D_Device))) return false;

   // One last check to be sure.
   if(D3D_Device == NULL) return false;

   // Set render states.
   D3D_Device->SetRenderState(D3DRS_LIGHTING, FALSE);
	D3D_Device->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
   D3D_Device->SetRenderState(D3DRS_ZENABLE, TRUE);

   // Init object.
   if(!InitializeObject()) return false;

   // Set the window perpective.
   D3DXMatrixPerspectiveFovLH(&Projection, 45.0f, (float)(640/480), 0.1f, 500.0f);
   D3D_Device->SetTransform(D3DTS_PROJECTION, &Projection);

   return true;
}


bool InitializeObject()
{
   // Seeing how our object is simple, we can just create each bone like so.
   // Our SetBone function takes the parent index and bone length.
   // SetBone create each bone for us.
   bones[0].SetBone(-1, 4.5f);
   bones[1].SetBone(0, 4.5f);
   
   // Set each of the model's vertex data.  SetVertex takes the vertex position, normal,
   // 4 bone index values, 4 bone weights, color in r, g, b, a, and a number of bones
   // that effect this.  The max should be 4 since 4 weights are used for skinning.
   modelPoints[0].SetVertex(CVector4(1.0f, 0.0f, 0.0f), CVector4(0.0f, 0.0f, 1.0),
                            0, 0, 0, 0, 1.0f, 0.0f, 0.0f, 0.0f, 0, 0, 255, 255, 1);
   modelPoints[1].SetVertex(CVector4(-1.0f, 0.0f, 0.0f), CVector4(0.0f, 0.0f, 1.0),
                            0, 0, 0, 0, 1.0f, 0.0f, 0.0f, 0.0f, 0, 0, 255, 255, 1);
   modelPoints[2].SetVertex(CVector4(-1.0f, 3.0f, 0.0f), CVector4(0.0f, 0.0f, 1.0),
                            0, 1, 0, 0, 0.50f, 0.50f, 0.0f, 0.0f, 0, 0, 255, 255, 2);
   modelPoints[3].SetVertex(CVector4(1.0f, 3.0f, 0.0f), CVector4(0.0f, 0.0f, 1.0),
                            0, 1, 0, 0, 0.50f, 0.50f, 0.0f, 0.0f, 0, 0, 255, 255, 2);

   modelPoints[4].SetVertex(CVector4(1.0f, 3.0f, 0.0f), CVector4(0.0f, 0.0f, 1.0),
                            0, 1, 0, 0, 0.50f, 0.50f, 0.0f, 0.0f, 0, 255, 0, 255, 2);
   modelPoints[5].SetVertex(CVector4(-1.0f, 3.0f, 0.0f), CVector4(0.0f, 0.0f, 1.0),
                            0, 1, 0, 0, 0.50f, 0.50f, 0.0f, 0.0f, 0, 255, 0, 255, 2);
   modelPoints[6].SetVertex(CVector4(-1.0f, 6.0f, 0.0f), CVector4(0.0f, 0.0f, 1.0),
                            0, 1, 0, 0, 0.35f, 0.65f, 0.0f, 0.0f, 0, 255, 0, 255, 2);
   modelPoints[7].SetVertex(CVector4(1.0f, 6.0f, 0.0f), CVector4(0.0f, 0.0f, 1.0),
                            0, 1, 0, 0, 0.35f, 0.65f, 0.0f, 0.0f, 0, 255, 0, 255, 2);

   modelPoints[8].SetVertex(CVector4(1.0f, 6.0f, 0.0f), CVector4(0.0f, 0.0f, 1.0),
                            0, 1, 0, 0, 0.35f, 0.65f, 0.0f, 0.0f, 255, 0, 0, 255, 2);
   modelPoints[9].SetVertex(CVector4(-1.0f, 6.0f, 0.0f), CVector4(0.0f, 0.0f, 1.0),
                            0, 1, 0, 0, 0.35f, 0.65f, 0.0f, 0.0f, 255, 0, 0, 255, 2);
   modelPoints[10].SetVertex(CVector4(-1.0f, 9.0f, 0.0f), CVector4(0.0f, 0.0f, 1.0),
                             1, 0, 0, 0, 1.0f, 0.0f, 0.0f, 0.0f, 255, 0, 0, 255, 1);
   modelPoints[11].SetVertex(CVector4(1.0f, 9.0f, 0.0f), CVector4(0.0f, 0.0f, 1.0),
                             1, 0, 0, 0, 1.0f, 0.0f, 0.0f, 0.0f, 255, 0, 0, 255, 1);

   return true;
}


void UpdateBones()
{
   // Temp matrices to hold rotations and translations.
   CMatrix4x4 rotTemp1, rotTemp2, tempRelative;

   // Here we will loop through the list of bones and update the skeleton.
   for(int i = 0; i < MAX_BONES; i++)
      {
         // Only the root bone will have a -1 value for its parent.  Root bones pretty much
         // are directly translated and rotated but it does not inherit anything.
         if(bones[i].parent == -1)
            {
               // Since this demo does not use keyframes, we will directly manipulate the
               // root bones translation/rotation.  The absolute of any root bone is its
               // relative.
               bones[i].relative.Translate(trans.x, trans.y, trans.z);
               bones[i].absolute = bones[i].relative;

               // For display purposes only.
               displayBones[i] = bones[i].relative;
            }
         else
            {
               // But if this is not the root bone, we have a little more work to do.

               bones[i].relative.Clear();

               // First we must move this bone into position from the parent.
               bones[i].relative.Translate(0, bones[bones[i].parent].length * i, 0);

               // Lets rotate the non-root bones with the mouse.  First
               // we will use temp matrices to hold both x and y rotations then we
               // will store them combined in the relative matrix of this bone.
               rotTemp1.Rotate(xRotation, 0, 1, 0);
               rotTemp2.Rotate(yRotation, 0, 0, 1);

               // Now since this is not a root bone we must do a little more to calculate the
               // child bones (and siblings).  What this means is that the absolute matrix
               // of this bone is equal to the absolute of the parent bone * the relative
               // matrix of this bone.  Since we store the offset inverse in a different
               // matrix, we use absolute = parent absolute * offset(relative without rotations)
               // * rotation * inverse offset(negative relative without rotations).

               // To do this we store a copy of our relative matrix then inverse that copy.
               tempRelative = bones[i].relative;
               tempRelative.inverseTranslate();

               // For display purposes only.
               displayBones[i] = bones[bones[i].parent].absolute * bones[i].relative *
                                 (rotTemp1 * rotTemp2);

               // Next we calculate the relative matrix by combining the translation and
               // rotations of this matrix.
               bones[i].relative = bones[i].relative * (rotTemp1 * rotTemp2) * tempRelative;
               
               // Compute the absoulte matrix as parent's absolute by this relative.
               bones[i].absolute = bones[bones[i].parent].absolute * bones[i].relative;
            }
      }
}


void RenderScene()
{
   // Clear the screen to black.
   D3D_Device->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                     D3DCOLOR_COLORVALUE(0.0f, 0.0f, 0.0f, 1.0f), 1.0f, 0);

   // Tell Direct 3D we are now going to start drawing.
   D3D_Device->BeginScene();

      // This will hold the view matrix.
      D3DXMATRIXA16 View;

      D3DXMatrixIdentity(&View);
      D3DXVECTOR3 Eye(0.0f, 4.0f, 15.0f);
      D3DXVECTOR3 LookAt(0.0f, 4.0f, 0.0f);
      D3DXVECTOR3 Up(0.0f, 1.0f, 0.0f);

      // Set view.
      D3DXMatrixLookAtLH(&View, &Eye, &LookAt, &Up);
      D3D_Device->SetTransform(D3DTS_VIEW, &View);

      // Load identity into world matrix.
      D3DXMATRIX identity;
      D3DXMatrixIdentity(&identity);
      D3D_Device->SetTransform(D3DTS_WORLD, (D3DMATRIX*)&identity);


      // Update the bone list by updating the absolute matrices of all bones.
      UpdateBones();
      
      // Temp matrix used to store the absolute matrix of a bone.
      CMatrix4x4 mat;

      // Used to store the displayable, animated mesh.
      Vertex animatedModel[MAX_MESH_POINTS];
      int index = 0;

      // Loop through the meshes and render each vertex.
      for(int j = 0; j < MAX_MESHES; j++)
      {
         for(int i = 0; i < MAX_MESH_POINTS; i++)
            {
               // Get a pointer this this vertex rather than having to type it out
               // the long way (modelPoints[i + j * MAX_MESH_POINTS].(whatever).
               // We do the same thing with mat so we can save source line space.
               CBoneVertex *thisVertex = &modelPoints[i + j * MAX_MESH_POINTS];

               // We will add to pos over time as we calculate the skinning weights.
               CVector4 pos;

               // Loop through all of the weights and add each ones contribution to the final
               // vertex position.  Basically we loop through each weight for each bone that can
               // influence this vertex and we add them up.  We multiply each position
               // by the bone matrix in question, then we multiply this by the weight of this
               // bone.  For the vertex pos we multiply it against the 4x4 matrix.
               for(int i2 = 0; i2 < thisVertex->numBones; i2++)
                  {
                     // Copy the abolute of the bone that influences this point.
                     // Not necessary but it makes the code on each line shorter below.
                     mat = bones[thisVertex->boneIndex[i2]].absolute;

                     // Add the contribution of this pos and normal against the bone.   
                     pos += (mat.VectorMatrixMultiply(thisVertex->pos) *
                            thisVertex->weights[i2]);
                  }

               // Set this vertex.
               animatedModel[index].x = pos.x;
               animatedModel[index].y = pos.y;
               animatedModel[index].z = pos.z;
               animatedModel[index].color = D3DCOLOR_XRGB(thisVertex->r,
                  thisVertex->g, thisVertex->b);
               index++;
            }

         // Draw out mesh.
         D3D_Device->SetFVF(D3DFVF_D3DVertex);
         D3D_Device->DrawPrimitiveUP(D3DPT_TRIANGLEFAN, 2, animatedModel, sizeof(Vertex));
         index = 0;
      }


      // Loop through and draw each bone.  This is for display purposes only just so
      // we can see the bones.
      for(int i = 0; i < MAX_BONES; i++)
         {
            D3D_Device->SetTransform(D3DTS_WORLD, (D3DMATRIX*)displayBones[i].matrix);
      
            // Draw a quad in lines.
            Vertex baseLines[] = { -0.4f, 0.0f, -0.4f, D3DCOLOR_XRGB(255,255,255),
               0.4f, 0.0f, -0.4f, D3DCOLOR_XRGB(255,255,255),
               0.4f, 0.0f, -0.4f, D3DCOLOR_XRGB(255,255,255),
               0.4f, 0.0f, 0.4f, D3DCOLOR_XRGB(255,255,255),
               0.4f, 0.0f, 0.4f, D3DCOLOR_XRGB(255,255,255),
               -0.4f, 0.0f, 0.4f, D3DCOLOR_XRGB(255,255,255),
               -0.4f, 0.0f, 0.4f, D3DCOLOR_XRGB(255,255,255),
               -0.4f, 0.0f, -0.4f, D3DCOLOR_XRGB(255,255,255) };

            // Now from each end point go up until the bone length.
            Vertex coneLines[] = { -0.4f, 0.0f, -0.4f, D3DCOLOR_XRGB(255,255,255),
               0.0f, bones[i].length, 0.0f, D3DCOLOR_XRGB(255,255,255),
               0.4f, 0.0f, -0.4f, D3DCOLOR_XRGB(255,255,255),
               0.0f, bones[i].length, 0.0f, D3DCOLOR_XRGB(255,255,255),
               0.4f, 0.0f, 0.4f, D3DCOLOR_XRGB(255,255,255),
               0.0f, bones[i].length, 0.0f, D3DCOLOR_XRGB(255,255,255),
               -0.4f, 0.0f, 0.4f, D3DCOLOR_XRGB(255,255,255),
               0.0f, bones[i].length, 0.0f, D3DCOLOR_XRGB(255,255,255), };

            D3D_Device->SetFVF(D3DFVF_D3DVertex);
            D3D_Device->DrawPrimitiveUP(D3DPT_LINELIST, 4, baseLines, sizeof(Vertex));
            D3D_Device->DrawPrimitiveUP(D3DPT_LINELIST, 4, coneLines, sizeof(Vertex));
         }

   // Done drawing.
   D3D_Device->EndScene();

   // Swap the old frame with the new one.
   D3D_Device->Present(NULL, NULL, NULL, NULL);
}


void ShutdownDirect3D()
{
   // Here we release the Direct3D objects.
   if(D3D_Device != NULL)
      {
         D3D_Device->Release();
         D3D_Device = NULL;
      }

   if(Direct3D_Object != NULL)
      {
         Direct3D_Object->Release();
         Direct3D_Object = NULL;
      }
}


// Recap:

// If you run this demo you will see a 2D set of quads being rotated with the mouse
// along with the bones.  Since this object is 2D it does not look as good as if it
// was a 3D mesh but trust me, it works.  Also keep in mind that the rotations are not
// limited so you can twist the object 360 degrees.  This is only the basics but it is a
// good start.  All we did in this demo was add 3 quads to the scene and allowed our
// bones to manipulate them.  In the next demo we will work with keyframe animation
// bone data.


// Copyright February 2005
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com