/*
   Class Name:

      CMatrix.

   Created by:

      Allen Sherrod (Programming Ace of www.UltimateGameProgramming.com).

   Description:

      This class represents 4x4 Matrix for use in OpenGL and Direct3D.
*/


#ifndef CMATRIX_H
#define CMATRIX_H
#include<string.h>
#include"CVector.h"


class CMatrix4x4
{
   public:
      CMatrix4x4();                                // Constructor.
      //CMatrix4x4(const CMatrix4x4 &m);             // Overloaded constructors.
      CMatrix4x4(float r11, float r12, float r13, float r14,
                 float r21, float r22, float r23, float r24,
                 float r31, float r32, float r33, float r34,
                 float r41, float r42, float r43, float r44);
      ~CMatrix4x4();                               // Destructor.

      void Clear();                                // Reset the matrix.
      void Zero();                                 // Reset the matrix to all 0s.

      void operator=(CMatrix4x4 &m);                // Overloaded = sign.
      CMatrix4x4 operator-(CMatrix4x4 m);          // Overloaded - sign.
      CMatrix4x4 operator+(CMatrix4x4 m);          // Overloaded + sign.
      CMatrix4x4 operator*(CMatrix4x4 m);          // Overloaded * sign.
      CMatrix4x4 operator/(CMatrix4x4 m);          // Overloaded / sign.

      void operator +=(CMatrix4x4 m);              // Overloaded += sign.
      void operator -=(CMatrix4x4 m);              // Overloaded -= sign.
      void operator *=(CMatrix4x4 m);              // Overloaded *= sign.
      void operator /=(CMatrix4x4 m);              // Overloaded /= sign.
      
      CMatrix4x4 operator-(float f);               // Overloaded - sign.
      CMatrix4x4 operator+(float f);               // Overloaded + sign.
      CMatrix4x4 operator*(float f);               // Overloaded * sign.
      CMatrix4x4 operator/(float f);               // Overloaded / sign.

      void operator -=(float f);                   // Overloaded -= sign.
      void operator +=(float f);                   // Overloaded += sign.
      void operator *=(float f);                   // Overloaded *= sign.
      void operator /=(float f);                   // Overloaded /= sign.

      bool operator ==(CMatrix4x4 v);              // Overloaded == sign.
      bool operator !=(CMatrix4x4 v);              // Overloaded != sign.

      void Translate(CVector4 Translate);          // Translate a point in 3D.
      void Translate(float x, float y, float z);   // Translate a point in 3D.
      void inverseTranslate();                     // Translate the opposite way.
      CVector4 inverseTranslateVector(CVector4 Translate);// Translate point inverse.
      void Rotate(double angle, float x, float y,
                  float z);                        // Rotate a 3D point.
      void RotateRadians(double x, double y, double z);// Rotate using radians.
      bool inverseMatrix(CMatrix4x4 m);            // Inverse the current matrix data.
      bool Transpose(CMatrix4x4 m);                // Invert a matrix.
      CVector4 VectorMatrixMultiply(CVector4 &v);   // Vector matrix multiply.
      CVector4 VectorMatrixMultiply3x3(CVector4 &v);// Vector matrix multiply 3x3 matrix.
      CVector4 inverseVectorMatrixMultiply3x3(CVector4 v);// Vector inverse multiply.
      bool CreateShadowMatrix(CVector4 planeNormal,
                              CVector4 lightPos);  // Create a shadow matrix.

      float matrix[16];                            // The 4x4 matrix in a 1D array.
};

#endif


// Copyright May 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com